#  Copyright (C) 2025 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - Example - Show random walk results
# ==========================================
#
#
# 
"""

import ararpy as ap
import numpy as np
import pdf_maker as pm
import os

import matplotlib
from matplotlib.collections import PathCollection

matplotlib.use('TkAgg')
matplotlib.rc('font',family='Arial', size=10)
import matplotlib.pyplot as plt
# 设置全局字体，确保中文正常显示
# plt.rcParams["font.family"] = ["SimHei"]  # 中文字体
# plt.rcParams['axes.unicode_minus'] = False  # 解决负号显示问题


main_color = ['#397DA1', '#BA5624', '#212121', '#6C5D1E', '#BC3D85', '#3C6933']  # blue red black
middle_color = ['#83CDFA', '#F1B595', '#737373', '#C0A737', '#E9CDE1', '#84B775']
shallow_color = ['#E0F1FE', '#FBEBE3', '#DDDDDD', '#F9E16F', '#CB73B0', '#B8F1A7']

colors =[
    '#397DA1', '#BA5624', '#212121', '#6C5D1E', '#BC3D85', '#3C6933',
    '#83CDFA', '#F1B595', '#737373', '#C0A737', '#E9CDE1', '#84B775',
    '#E0F1FE', '#FBEBE3', '#DDDDDD', '#F9E16F', '#CB73B0', '#B8F1A7',
]


def export_to_pdf(axs, filename="unknown"):

    params_list = {
        "page_size": 'a4', "ppi": 72, "width": 9.5, "height": 6,
        "pt_width": 0.8, "pt_height": 0.8, "pt_left": 0.16, "pt_bottom": 0.18,
        "offset_top": 0, "offset_right": 0, "offset_bottom": 20, "offset_left": 30,
        "plot_together": False, "show_frame": False,
        'xlabel_offset': 8, 'ylabel_offset': 2
    }

    plot_data = {
        "data": [transform(axs)],
        "file_name": filename,
        "plot_names": [f"plotname"],
    }

    filepath = os.path.join(r"C:\Users\Young\Downloads", f"{filename}.pdf")
    cvs = [[ap.smp.export.get_cv_from_dict(plot, **params_list) for plot in plot_data['data']]]
    for i in range(len(cvs[0])):
        pt = cvs[0][i]._plot_areas[0]
        title = pt.get_comp(comp_name="title")
        title._y -= 2
        title._z_index = 999
        pt.text(50, title._y, text=f"({['a', 'b', 'c', 'd'][i]})", coordinate='pt', clip=False, size=8, z_index=299, v_align='top')
        for comp in pt._components:
            if isinstance(comp, pm.Scatter):
                comp._type = 'rec'
                comp._size = 1.5
        for index, legned in enumerate(list(filter(lambda cp: cp.name() == 'legend', pt._components))):
            legned._size = 7
            legned._z_index = 250
            legned._h_align = "left"
            legned._v_align = "center"
            if i == 0:  # age spectra
                legned._x = 175
                legned._y = 40 + index * 10
            elif i == 1:  # cooling history
                legned._x = 65
                legned._y = 135 - index * 10
            elif i == 2:  # arrhenius
                legned._x = 75
                legned._y = 40 + index * 10
            else:
                legned._x = 65
                legned._y = 40 + index * 10
            for comp in pt._components:
                if legned._text in comp.name() and "legend" in comp.name():
                    comp._z_index = 250
                    if isinstance(comp, pm.Scatter):
                        comp._x = legned._x - 10
                        comp._y = legned._y
                    if isinstance(comp, pm.Line):
                        comp._start = [legned._x - 16, legned._y]
                        comp._end = [legned._x - 4, legned._y]


    filepath = ap.smp.export.export_chart_to_pdf(cvs, filename, filepath)

    return filepath


def read_sample(arr_file_path):
    # 读取样品信息
    if not arr_file_path.endswith('.arr'):
        for root, dirs, files in os.walk(arr_file_path):
            for file in files:
                if file.endswith('.arr'):
                    arr_file_path = os.path.join(arr_file_path, file)
                    break
    print(f"arr file: {arr_file_path}")
    sample = ap.from_arr(file_path=arr_file_path)
    name = sample.name()
    sequence = sample.sequence()
    nsteps = sequence.size
    te = np.array(sample.TotalParam[124], dtype=np.float64)
    ti = (np.array(sample.TotalParam[123], dtype=np.float64) / 60).round(2)  # time in minute
    nindex = {"40": 24, "39": 20, "38": 10, "37": 8, "36": 0}
    argon = "39"
    if argon in list(nindex.keys()):
        ar = np.array(sample.DegasValues[nindex[argon]], dtype=np.float64)  # 20-21 Argon
        sar = np.array(sample.DegasValues[nindex[argon] + 1], dtype=np.float64)
    elif argon == 'total':
        all_ar = np.array(sample.CorrectedValues, dtype=np.float64)  # 20-21 Argon
        ar, sar = ap.calc.arr.add(*all_ar.reshape(5, 2, len(all_ar[0])))
        ar = np.array(ar)
        sar = np.array(sar)
    else:
        raise KeyError
    age = np.array(sample.ApparentAgeValues[2], dtype=np.float64)  # 2-3 age
    sage = np.array(sample.ApparentAgeValues[3], dtype=np.float64)
    f = np.cumsum(ar) / ar.sum()

    # 组合data
    dr2 = [1 for i in range(nsteps)]
    ln_dr2 = [1 for i in range(nsteps)]
    wt = [1 for i in range(nsteps)]
    data = np.array([sequence.value, te, ti, age, sage, ar, sar, f, dr2, ln_dr2, wt]).tolist()
    data.insert(0, (np.where(np.array(data[3]) > 0, True, False) & np.isfinite(data[3])).tolist())
    data.insert(1, [1 for i in range(nsteps)])
    for row in ap.calc.arr.transpose(data):
        print(row)
    return data, name


def read_ads(ads_file_path):
    released = []
    release_name = []

    ads_released = []
    index = 1
    if os.path.isdir(ads_file_path):
        for (dirpath, dirnames, fs) in os.walk(ads_file_path):
            for f in fs:
                if f.endswith(".ads"):
                    file_path = os.path.join(ads_file_path, f)
                    if not os.path.exists(file_path):
                        continue
                    # if "k=100" not in f:
                    #     continue
                    index += 1
                    release_name.append(f"Released{index}: {f}")
                    diff = ap.thermo.arw.read_ads(file_path)
                    print(f"{f = }, {len(diff.released_per_step) = }, {diff.atom_density = :.0e}")
                    print(f"{'kJ/mol, '.join([str(dom.energy / 1000) for dom in diff.domains])}")
                    print(f"{'kcal/mol, '.join([str(dom.energy / 4.181 / 1000) for dom in diff.domains])}")
                    ads_released.append(np.array(diff.released_per_step) / diff.natoms)


    # for i in range(len(ads_released)):
    #     released.append([i + 1, sum(ar[0:i + 1])  / sum(ar), *ads_released[i]])
    #
    # spectra_data.append(released)

    return ads_released, release_name


def plot_diff(loc, arr_file_path=None, argon=39):
    ads_released, release_name = read_ads(loc)

    fig, axs = plt.subplots(1, 1, figsize=(12, 8))

    index = 0
    for index, each_line in enumerate(ads_released):
        axs.plot(list(range(1, len(each_line) + 1)), each_line, c=colors[index], linewidth=1, label=release_name[index])

    if arr_file_path is not None:
        sample = ap.from_arr(file_path=arr_file_path)
        nindex = {"40": 24, "39": 20, "38": 10, "37": 8, "36": 0}
        argon = str(argon)
        if argon in list(nindex.keys()):
            ar = np.array(sample.DegasValues[nindex[argon]], dtype=np.float64)  # 20-21 Argon
            sar = np.array(sample.DegasValues[nindex[argon] + 1], dtype=np.float64)
        elif argon == 'total':
            all_ar = np.array(sample.CorrectedValues, dtype=np.float64)  # 20-21 Argon
            ar, sar = ap.calc.arr.add(*all_ar.reshape(5, 2, len(all_ar[0])))
            ar = np.array(ar)
            sar = np.array(sar)
        else:
            raise KeyError
        axs.plot(list(range(1, len(ar) + 1)), np.cumsum(ar) / np.sum(ar), c=colors[index+1], linewidth=1, label=f"{sample.name()}")

    axs.set_title(f'Ads released', loc='center', y=1,)
    axs.set_xlabel(f'Steps',)
    axs.set_ylabel(f'Aumulative Argon Released (100%)',)
    axs.legend(loc='lower right')
    fig.tight_layout()
    plt.show()

    export_to_pdf(axs, filename="test_ads")


def plot_spectra(loc, arr_file_path=None, argon=39, loc39=""):

    ar = []
    name = ""
    if arr_file_path is not None:
        sample = ap.from_arr(file_path=arr_file_path)
        name = sample.name()
        ar39 = np.array(sample.DegasValues[20], dtype=np.float64)  # 20-21 Argon
        ar40 = np.array(sample.DegasValues[24], dtype=np.float64)
        print(f"{ar40 = }")



    ads_released_ar40, release_name = read_ads(loc)
    ads_released_ar39, release_name = read_ads(loc39)

    fig, axs = plt.subplots(1, 1, figsize=(12, 8))

    index = 0
    for index, each_line in enumerate(ads_released_ar40):
        _ = np.array(each_line) * sum(ar40)
        ar40_model = [_[0]]
        for i in range(1, len(_)):
            ar40_model.append(_[i] - _[i - 1])

        _ = np.array(ads_released_ar39[0]) * sum(ar39)
        ar39_model = [_[0]]
        for i in range(1, len(_)):
            ar39_model.append(_[i] - _[i - 1])
        ar39_model = ar39_model[:58]
        print(f"{len(ar39_model) = }, {len(ar40_model) = }")

        x, y1, y2 = ap.calc.spectra.get_data(np.array(ar40_model)/np.array(ar39_model), np.zeros(len(ar39_model)), ar39_model)
        axs.plot(x, y1, c=colors[index], linewidth=1, label=release_name[index])

    x, y1, y2 = ap.calc.spectra.get_data(ar40/ar39, np.zeros(len(ar40)), ar39)
    axs.plot(x, y1, c=colors[index+1], linewidth=1, label=f"{name}")

    axs.set_title(f'Ads spectra', loc='center', y=1,)
    axs.set_xlabel(f'Steps',)
    axs.set_ylabel(f'Aumulative Argon Released (100%)',)
    axs.set_ylim(0, 8)
    axs.legend(loc='lower right')
    fig.tight_layout()
    plt.show()

    export_to_pdf(axs, filename="test_ads")



def transform(ax: plt.Axes):
    xlabels = [i.get_text().replace('−', '-') for i in ax.get_xticklabels()]
    ylabels = [i.get_text().replace('−', '-') for i in ax.get_yticklabels()]
    linestyles = {'-': 'solid', '--': 'dashed', '-.': 'dashdot', ':': 'dotted'}

    series = []
    for i, line in enumerate(ax.lines):
        xy_data = line.get_xydata()  # [[x1, y1], [x2, y2], ...]
        line_style = linestyles.get(line.get_linestyle(), 'solid')
        series.append({
            'type': 'series.line', 'id': f'line-{i}', 'name': f'line-{i}',
            'color': line.get_color(), 'line_width': 1, 'line_style': line_style,
            'data': xy_data, 'line_caps': 'none'
        })
        if bool(line._marker):
            series.append({
                'type': 'series.scatter', 'id': f'line-marker-{i}', 'name': f'line-marker-{i}',
                'stroke_color': line.get_markeredgecolor(), 'fill_color': line.get_markerfacecolor(),
                'data': xy_data, 'size': 2,
                # 'symbol': line._marker.markers.get(line.get_marker(), 'square'),
                'symbol': 'rec'
            })
    for i, collection in enumerate(ax.collections):
        series.append({
            'type': 'series.scatter', 'id': f'scatter-{i}', 'name': f'{collection.get_label()}',
            'stroke_color': collection.get_edgecolor()[0][:3], 'fill_color': collection.get_edgecolor()[0][:3],
            'data': collection.get_offsets(), 'size': 2,
            'symbol': 'rec'
        })

    for i, text in enumerate(ax.texts):
        xy_data = text.get_position()  # [[x1, y1], [x2, y2], ...]
        series.append({
            'type': 'series.text', 'id': f'text-{i}', 'name': f'text-{i}',
            'color': text.get_color(), 'data': [xy_data], 'text': text.get_text().replace('\n', '<r>'),
            'size': 8
        })

    series.append({
        'type': 'series.text', 'id': f'title', 'name': f'title',
        'color': 'black', 'data': [[sum(ax.get_xlim()) / 2, ax.get_ylim()[1]]],
        'h_align': "middle", 'v_align': "top",
        'text': ax.get_title(), 'size': 8
    })

    if ax.legend_ is not None:
        for handle, text in zip(ax.legend_.legend_handles, ax.legend_.texts):
            series.append({
                'type': 'series.text', 'id': f'legend', 'name': f'legend',
                'color': text.get_color(), 'data': [[ax.get_xlim()[0], ax.get_ylim()[0]]],
                'h_align': "left", 'v_align': "bottom",
                'text': text.get_text(), 'size': 8
            })
            if isinstance(handle, plt.Line2D):
                series.append({
                    'type': 'series.line', 'id': f'legend-line', 'name': f'legend-line-{text.get_text()}',
                    'color': handle.get_color(), 'data':[[ax.get_xlim()[0], ax.get_ylim()[0]], [ax.get_xlim()[1], ax.get_ylim()[1]]],
                    'line_width': 1, 'line_style': linestyles.get(handle.get_linestyle(), 'solid')
                })
            if isinstance(handle, PathCollection):
                stroke_c = handle.get_edgecolor()[0][:3]
                stroke_c = f"#{int(stroke_c[0]*255):02x}{int(stroke_c[1]*255):02x}{int(stroke_c[2]*255):02x}"
                fill_c = handle.get_facecolor()[0][:3]
                fill_c = f"#{int(fill_c[0]*255):02x}{int(fill_c[1]*255):02x}{int(fill_c[2]*255):02x}"
                series.append({
                    'type': 'series.scatter', 'id': f'legend-scatter', 'name': f'legend-scatter-{text.get_text()}',
                    'stroke_color': stroke_c, 'fill_color': fill_c,
                    'data': [[sum(ax.get_xlim()) / 2, sum(ax.get_ylim()) / 2]],
                    'size': 2, 'symbol': 'rec'
                })

    data = {
        'xAxis': [{
            'extent': ax.get_xlim(), 'interval': xlabels, 'title': ax.get_xlabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'yAxis': [{
            'extent': ax.get_ylim(), 'interval': ylabels, 'title': ax.get_ylabel(),
            'nameLocation': 'middle', 'show_frame': True, 'label_size': 8, 'title_size': 8,
        }],
        'series': series
    }

    # print(data)
    return data


if __name__ == "__main__":
    loc = r"D:\DjangoProjects\webarar\private\mdd\20240920_24FY88a\1doms"
    loc = r"D:\DjangoProjects\webarar\private\mdd\20240920_24FY88a\2doms"
    loc = r"D:\DjangoProjects\webarar\private\mdd\20240920_24FY88a\5doms"
    loc = r"D:\DjangoProjects\webarar\private\mdd\20240920_24FY88a\6doms"
    # loc = r"D:\DjangoProjects\webarar\private\mdd\24FY88a-Ar40\thermo-history-0628"
    # loc = r"D:\DjangoProjects\webarar\private\mdd\24FY88a-Ar40\thermo-history-0628\heating_experiment"
    loc = r"D:\DjangoProjects\webarar\private\mdd\24FY88a-Ar40\thermo-history-0919\lab"
    loc2 = r"D:\DjangoProjects\webarar\private\mdd\24FY88a-Ar40\thermo-history-0919\ar39"
    # loc = r"D:\DjangoProjects\webarar\private\mdd\24FY88a-Ar40\thermo-history-0919\ar39"
    # arr_file_path = r"D:\DjangoProjects\webarar\private\mdd\20240920_24FY88a\20240920_24FY88a.arr"
    arr_file_path = r"D:\DjangoProjects\webarar\private\mdd\24FY88a-Ar40\20240920_24FY88a-smooth.arr"
    # plot_diff(loc, arr_file_path, argon=39)
    plot_spectra(loc, arr_file_path, argon=40, loc39=loc2)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - arr
# ==========================================
#
# Basic functions related to array-like objects
# List, np.array, pd.DataFrame
#
"""

# === Internal import ===
import traceback

from . import err
# === External import ===
import numpy as np
from scipy.stats import distributions
from types import MethodType
from typing import Optional, List, Union
from decimal import Decimal


# =======================
# Error propagation for array-like objects
# =======================
def mul(*args):
    """
    Parameters
    ----------
    args :
        2D array like objects of numbers
        lists: [[1, 2, 3], [1, 2, 3]], [[3, 4, 5], [3, 4, 5]]
        np.array([[1, 2, 3], [1, 2, 3]]), np.array([[3, 4, 5], [3, 4, 5]])
        pd.DataFrame([[1, 1], [2, 2], [3, 3]]), pd.DataFrame([[3, 3], [4, 4], [5, 5]])

    Returns
    -------
    two-D list, [values, errors]
    """
    v = np.array(args)
    res = np.zeros([2, v.shape[-1]], dtype=np.float64)
    for i in range(v.shape[-1]):
        res[0, i] = np.prod(v[:, 0, i])
        res[1, i] = err.mul(*v[:, :, i])

    return res


def div(a, b):
    """
    Y = a / b
    Parameters
    ----------
    a : two dimensional array like objects of numbers
    b : two dimensional array like objects of numbers
        lists: [[1, 2, 3], [1, 2, 3]], [[3, 4, 5], [3, 4, 5]]
        np.array([[1, 2, 3], [1, 2, 3]]), np.array([[3, 4, 5], [3, 4, 5]])
        pd.DataFrame([[1, 1], [2, 2], [3, 3]]), pd.DataFrame([[3, 3], [4, 4], [5, 5]])

    Returns
    -------
    [values, errors]
    """
    n = np.shape([a, b])[-1]
    k0, k1 = [], []
    for i in range(n):
        try:
            k0.append(np.divide(a[0][i], b[0][i]))
            k1.append((np.divide(a[1][i] ** 2, b[0][i] ** 2) + np.divide(a[0][i] ** 2 * b[1][i] ** 2, b[0][i] ** 4)) ** .5)
        except TypeError as e:
            raise TypeError(f"a = {a[0][i]}, b = {b[0][i]},  {e}")
    return [k0, k1]


def rec(a):
    """
    Parameters
    ----------
    a :
        2D array like objects of numbers

    Returns
    -------
        two dimensional list, [values, errors]
    """
    try:
        n = np.shape(a)[1]
        k0, k1 = [], []
        for i in range(n):
            k0.append(np.divide(1, a[0][i]))
            k1.append(err.rec((a[0][i], a[1][i])))
        return [k0, k1]
    except IndexError:
        return [1 / a[0], err.rec(a)]


def add(*args):
    """ For y = x1 + x2 + ...
    Parameters
    ----------
    args : two dimensional lists, or tuple of two list
        [x1, sx1], [x2, sx2], ...

    Returns
    -------
    list
    """
    n = np.shape(args)[-1]
    k0, k1 = [], []
    for i in range(n):
        k0.append(sum([arg[0][i] for arg in args]))
        k1.append(err.add(*[arg[1][i] for arg in args]))
    return [k0, k1]


def sub(*args):
    """ For y = x1 - x2 - ...
    Parameters
    ----------
    args : two dimensional lists, or tuple of two list
        [x1, sx1], [x2, sx2], ...

    Returns
    -------

    """
    args = np.array(args, dtype=np.float64)
    args[np.isnan(args)] = 0
    n = np.shape(args)[-1]
    k0, k1 = [], []
    for i in range(n):
        k0.append(sum([arg[0][i] * [-1, 1][index == 0] for index, arg in enumerate(args)]))
        k1.append(err.add(*[arg[1][i] for arg in args]))
    return [k0, k1]


def cor(sx: list, sy: list, sz: list):
    """
    Parameters
    ----------
    sx
    sy
    sz

    Returns
    -------

    """
    n = np.shape([sx, sy, sz])[-1]
    return [err.cor(sx[i], sy[i], sz[i]) for i in range(n)]


def mul_factor(a, factor, isRelative: bool = False):
    """ a × factor
    Parameters
    ----------
    a :
    factor : two dimensional list
    isRelative : if the error of factor is relative

    Returns
    -------

    """
    f, sf = np.array(factor)
    if isRelative:
        sf = f * sf / 100
    return mul(a, (f, sf))


def rec_factor(factor, isRelative: bool = False):
    """ 1 / factor
    Parameters
    ----------
    factor : two dimensional list
    isRelative : if the error of factor is relative

    Returns
    -------

    """
    f, sf = np.array(factor)
    if isRelative:
        sf = f * sf * 100
    return rec((f, sf))


# =======================
# Basic functions for array-like objects
# =======================
def partial(a: list, rows=None, cols=None):
    """ Get partial object from a two dimensional list according to the given rows and cols.

    Parameters
    ----------
    a : two dimensional array-like object with shape of (m, n)
    rows :
        rows index like [1, 2, ...], or int
    cols :
        cols index like [0 ,1, ...], or int

    Returns
    -------
    array-like object, if rows or cols is int, one dimensional list will be returned
    """
    res = []
    try:
        (m, n) = np.shape(a)
    except ValueError:
        # ValueError, the requested array has an inhomogeneous shape
        return partial(homo(a), rows=rows, cols=cols)
    default_col = [np.nan for i in range(n)]
    if rows is None:
        rows = list(range(n))
    if cols is None:
        cols = list(range(m))
    if isinstance(cols, list):
        res = [a[i] if i != -1 else default_col for i in cols]
    elif isinstance(cols, int):
        res = a[cols] if cols != -1 else default_col
    if isinstance(rows, list):
        if isinstance(cols, list):
            res = [[i[j] for j in rows] for i in res]
        else:
            res = [res[i] for i in rows]
    elif isinstance(rows, int):
        res = res[rows]
    return res


def array_as_float(p_object):
    """ Get array of floats from list like object
    Parameters
    ----------
    p_object

    Returns
    -------

    """
    try:
        return np.array(p_object, dtype=np.float64)
    except ValueError as e:
        item = str(e).split('could not convert string to float: ')
        return array_as_float(replace(p_object, item[1][1:-1], np.nan))


def replace(array, old, new):
    """ Replace all old items with new items in a array

    Parameters
    ----------
    array
    old : function or object
    new

    Returns
    -------

    """
    if np.iterable(array) and not isinstance(array, str):
        res = [replace(i, old, new) for i in array]
    else:
        if callable(old):
            if old(array):
                res = new
            else:
                res = old
        elif array in [old]:
            res = new
        else:
            res = array
    return res


def remove(array, old):
    """ Remove all old items

    Parameters
    ----------
    array : n dimensional list
    old : tuple, like (None, ), (None, np.nan)

    Returns
    -------

    """
    if is_oneD(array):
        res = []
        for each in array:
            if each in old:
                continue
            res.append(each)
    else:
        res = [remove(each, old=old) for each in array]
    return res


def merge(*args):
    """ Merge list objects, one dimensional or two dimensional list should be passed in.

    Parameters
    ----------
    args : one or two dimensional list

    Returns
    -------

    """
    res = []
    for arg in args:
        if is_oneD(arg):
            res.append(arg)
        else:
            res = res + arg
    return res


def multi_append(a, *args):
    """
    Parameters
    ----------
    a
    args

    Returns
    -------

    """
    for arg in args:
        a.append(arg)
    return a


def filter(a: list, func: MethodType, get: Optional[MethodType], unique: Optional[bool]):
    """
    Parameters
    ----------
    a
    func
    get
    unique

    Returns
    -------

    """
    res = []
    for i in a:
        if func(i):
            if type(get) is MethodType:
                res.append(get(i))
            else:
                res.append(i)
            if unique:
                return res[0]
    return res


def create_arr(shape: tuple):
    """
    Parameters
    ----------
    shape

    Returns
    -------

    """
    if len(shape) == 1:
        return []
    return [create_arr(shape[1:]) for i in range(shape[0])]


def is_empty(a):
    """ Check if a list is empty. Return True if it is empty, or it is not a list,
    or all items in it are '', None or np.nan.

    Parameters
    ----------
    a : list

    Returns
    -------
    bool. True if a is empty or is not a instance of list
    """
    if not isinstance(a, list) or a == [] or len(a) == 0:
        return True
    for i in a:
        if i not in ['', None, np.nan]:
            return False
    return True


def ndim(obj):
    """

    Parameters
    ----------
    obj

    Returns
    -------

    """
    if np.iterable(obj) and not isinstance(obj, str):
        if len(obj) == 0:
            return 1
        return max([ndim(i) for i in obj]) + 1
    return 0


def is_oneD(obj):
    """ Check if a list-like object is one dimensional
    Parameters
    ----------
    obj

    Returns
    -------

    """
    try:
        return ndim(obj) == 1
    except (ValueError, AttributeError):
        return False


def is_twoD(obj):
    """ Check if a list-like object is two dimensional
    Parameters
    ----------
    obj

    Returns
    -------

    """
    try:
        return ndim(obj) == 2
    except (ValueError, AttributeError):
        return False


def is_homo(obj):
    """ Check if the requested obj is homogeneous

    Parameters
    ----------
    obj

    Returns
    -------

    """
    try:
        np.shape(obj)
    except (ValueError, AttributeError):
        return False
    else:
        return True


def homo(obj):
    """ Get a homogeneous two dimensional list. If the passed list object is inhomogeneous,
    np.nan will be added to make it homogeneous.

    Parameters
    ----------
    obj : two dimensional list

    Returns
    -------

    """
    if not is_twoD(obj):
        return obj
    try:
        np.shape(obj)
    except (ValueError, AttributeError):
        pass
    else:
        return obj
    length = max(len(i) for i in obj)
    return [i if len(i) == length else i + [np.nan] * (length - len(i)) for i in obj]


def transpose(obj, ignore: bool = True):
    """ Get the transpose of a list-like object
    Parameters
    ----------
    obj : two dimensional list-like object
    ignore : return obj if ignore error is true

    Returns
    -------

    """
    try:
        if len(obj) == 0:
            return obj
        if not is_twoD(obj):
            raise ValueError("The requested object is not two dimensional.")
        obj = obj if is_homo(obj) else homo(obj)
        m, n = np.shape(obj)
        return [[obj[i][j] for i in range(m)] for j in range(n)]
    except (ValueError, TypeError):
        print(traceback.format_exc())
        if ignore:
            return obj
        else:
            raise ValueError


def remove_empty(a: list):
    index = 0
    for i in range(len(a)):
        if not is_empty(a[-(i + 1)]):
            index = len(a) - i
            break
    return transpose(a[:index])


def get_item(obj: list, loc: (list, tuple, int), default: Union[str, int, float, bool] = None,
             base: Union[int, tuple, list] = 0) -> Union[str, int, float, bool]:
    """ Get item from a n-dimension list
    Parameters
    ----------
    obj
    loc
    default
    base

    Returns
    -------
    """
    try:
        if not isinstance(obj, list):
            raise TypeError
        if isinstance(loc, int):
            loc = [loc]
        if isinstance(base, int):
            base = [base for i in range(len(loc))]
        if len(loc) == 1:
            return obj[loc[0] - base[0]] if loc[0] - base[0] >= 0 else ""
        if loc[0] - base[0] < 0:
            raise IndexError
        return get_item(obj[loc[0] - base[0]], loc[1:], base=base[1:], default=default)
    except (IndexError, TypeError, ValueError):
        return default
    except:
        raise ValueError(f"Cannot get a item from {obj = }, {loc = }")


# =======================
# Weighted mean
# =======================
def wtd_mean(a: list, e: list, sf: int = 1, adjust_error: bool = True):
    """
    :param a: x
    :param e: error
    :param sf: sigma number for age error, default = 1
    :param adjust_error: adjust error by multiply Sqrt(MSWD)
    :return: error-weighted mean value | error in 1 sigma | number of data points | MSWD | Chisq | P value
    """
    a, e = np.array([a, e])
    e = e / sf  # change error to 1 sigma
    k2 = a.size
    df = k2 - 1
    if k2 == 0:
        return np.nan, np.nan, 0, np.nan, np.nan, np.nan
    wt = 1 / e ** 2
    k0 = sum(a * wt) / sum(wt)  # error weighting
    k4 = sum((a - k0) ** 2 * wt)  # Chi square
    k3 = k4 / df  # MSWD mentioned in Min et al., 2000
    if adjust_error and k3 > 1:
        k1 = (k3 / sum(wt)) ** .5
    else:
        k1 = (1 / sum(wt)) ** .5
    k5 = distributions.chi2.sf(k4, df)
    return k0, k1, k2, k3, k4, k5


# =======================
# Number format
# =======================
def change_number_format(numbers: Union[list, tuple], flag: str = "Scientific", precision: int = None):
    """
    Parameters
    ----------
    precision
    flag
    numbers

    Returns
    -------

    """
    flag = flag.replace(" ", "")
    _precision = [len(str(num).split(".")[1]) if "." in str(num) else 0 for num in numbers]
    if flag.startswith("AddZero"):
        precision = max(_precision) if precision is None else precision
        for index, num in enumerate(numbers):
            numbers[index] = str(num) + ('.' if '.' not in str(num) else '') + '0' * (precision - _precision[index])
        return change_number_format(numbers, flag=flag.replace("AddZero", ""), precision=precision)

    if flag.startswith("Scientific"):
        numbers = [
            ("{:." + f"{(len(str(num).replace('.', '')) - 1) if precision is None else precision}" + "e}").format(
                Decimal(str(num))) for num in numbers]
        return change_number_format(numbers, flag=flag.replace("Scientific", ""), precision=precision)

    if flag.startswith("NonZero"):
        numbers = [str(num).rstrip("0") if "." in str(num) else str(num) for num in numbers]
        return change_number_format(numbers, flag=flag.replace("NonZero", ""), precision=precision)

    if flag.startswith("Round"):
        numbers = [str(round(float(num), min(_precision) if precision is None else precision)) for num in numbers]
        return change_number_format(numbers, flag=flag.replace("Round", ""), precision=precision)

    return numbers


if __name__ == '__main__':
    pass




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2025 Yang 
# ararpy - basic
# ==========================================
#
#
# 
"""
import numpy as np
from math import log

SEC2YEAR = 3600 * 24 * 365.2425
CAL2JOULE = 4.184
CM2UM = 10000
GAS_CONSTANT = 8.31446261815324  # in J / (mol K)


def gammq(a, x):
    if x < 0 or a <= 0:
        raise ValueError("ERROR(GAMMQ): x < 0 or a <= 0")

    if x < a + 1:
        gamser, gln = gser(a, x)
        return 1.0 - gamser
    else:
        gammcf, gln = gcf(a, x)
        return gammcf

def gser(a, x, itmax=100, eps=3e-7):
    gln = gammln(a)
    if x <= 0:
        if x < 0:
            raise ValueError("ERROR(GSER): x < 0")
        return 0.0, gln

    ap = a
    sum_ = 1.0 / a
    delta = sum_
    for n in range(1, itmax + 1):
        ap += 1
        delta *= x / ap
        sum_ += delta
        if abs(delta) < abs(sum_) * eps:
            break
    else:
        raise RuntimeError("ERROR(GSER): a too large, itmax too small")

    gamser = sum_ * np.exp(-x + a * np.log(x) - gln)
    return gamser, gln

def gcf(a, x, itmax=100, eps=3e-7):
    gln = gammln(a)
    gold = 0.0
    a0 = 1.0
    a1 = x
    b0 = 0.0
    b1 = 1.0
    fac = 1.0
    for n in range(1, itmax + 1):
        an = float(n)
        ana = an - a
        a0 = (a1 + a0 * ana) * fac
        b0 = (b1 + b0 * ana) * fac
        anf = an * fac
        a1 = x * a0 + anf * a1
        b1 = x * b0 + anf * b1
        if a1 != 0:
            fac = 1.0 / a1
            g = b1 * fac
            if abs((g - gold) / g) < eps:
                return g * np.exp(-x + a * np.log(x) - gln), gln
            gold = g
    else:
        raise RuntimeError("ERROR(GCF): a too large, itmax too small")

def gammln(xx):
    cof = np.array([76.18009173, -86.50532033, 24.01409822,
                    -1.231739516, 0.00120858003, -0.00000536382])
    stp = 2.50662827465
    half = 0.5
    one = 1.0
    fpf = 5.5

    x = xx - one
    tmp = x + fpf
    tmp = (x + half) * np.log(tmp) - tmp
    ser = one
    for j in range(6):
        x += one
        ser += cof[j] / x

    return tmp + np.log(stp * ser)


def fit(x, y, sigx, sigy, pho=None):
    """
    equivalent to York2 regression
    y = a + bx

    Parameters
    ----------
    x
    y
    sigx
    sigy
    pho

    Returns
    -------
    a: intercept
    b: slope
    siga: intercept uncertainty
    sigb: slope uncertainty
    chi2: chi-square value
    q: the incomplete gamma function value, should be large for the best fitting.

    """

    print(f"{x = }")
    print(f"{y = }")
    print(f"{sigx = }")
    print(f"{sigy = }")
    print(f"{pho = }")

    ndata = len(x)
    if pho is None:
        pho = np.zeros(ndata)
    nd = 20
    imax = 20
    xerr = 0.001
    wt = []
    a = 0
    b = -1.
    siga = 0
    sigb = 0
    chi2 = 0
    q = 0
    iter = 0

    while iter <= imax:
        sx = 0.
        sy = 0.
        st2 = 0.
        st3 = 0.
        ss = 0.
        b0 = b

        for i in range(ndata):
            wt.append(0)
            wt[i] = 1. / (sigy[i] ** 2 + b ** 2 * sigx[i] ** 2 - 2 * pho[i] * sigx[i] * sigy[i])
            ss = ss + wt[i]
            sx = sx + x[i] * wt[i]
            sy = sy + y[i] * wt[i]

            # print(f"{x[i] = }, {y[i] = }, {wt[i] = }")

        sxoss = sx / ss
        syoss = sy / ss

        for i in range(ndata):
            t1 = (x[i] - sxoss) * sigy[i] ** 2
            t2 = (y[i] - syoss) * sigx[i] ** 2 * b
            t3 = sigx[i] * sigy[i] * pho[i]
            st2 = st2 + wt[i] ** 2 * (y[i] - syoss) * (t1 + t2 - t3 * (y[i] - syoss))
            st3 = st3 + wt[i] ** 2 * (x[i] - sxoss) * (t1 + t2 - b * t3 * (x[i] - sxoss))

        b = st2 / st3
        iter = iter + 1

        # print(f"{sxoss = }, {syoss = }, {b = }, {abs(b0 - b) = }")

        if abs(b0 - b) > xerr:
            continue

        a = (syoss - sxoss * b)
        # print(f"{a = }, {b = }")
        sgt1 = 0.
        sgt2 = 0.

        for i in range(ndata):
            sgt1 = sgt1 + wt[i] * (x[i] - sxoss) ** 2
            sgt2 = sgt2 + wt[i] * x[i] ** 2

        sigb = (1. / sgt1) ** 0.5
        siga = sigb * (sgt2 / ss) ** 0.5
        chi2 = 0.

        for i in range(ndata):
            chi2 = chi2 + wt[i] * (y[i] - a - b * x[i]) ** 2

        q = gammq(0.5 * (ndata - 2), 0.5 * chi2)

        if abs(b0 - b) <= xerr:
            break

    return a, b, siga, sigb, chi2, q


def get_tc(da2, sda2, E, sE, cooling_rate=10, temp_in_celsius=True, temp=None, A: float = 27.0, R=None, pho=0):
    """
    E / (R Tc) = ln(A t D0/a2)
    E: diffusion activation energy
    R: gas constant, 8.314 J/(K*mol)
    Tc: closure temperature
    A: geometric constant, A is 55, 27, or 8.7 for volume diffusion from a sphere, cylinder or plane sheet respectively
    tau: time constant, can be determined by
        (1) a pre-defined cooling rate following t = -R Tc2 / E * dT / dt, where dT / dt is the average cooling rate of the local geological body
        (2) the age of dated sample following t = R * age * E * (Tp - Tc), where Tp is the temperature at present that could be 300 K.
    D0: the diffusion coefficient at the given temperature, which will be the frequency factor (the pre-exponential factor) whe the temperature is infinite high
    a: the effective diffusion radius, the characteristic dimension of the system, or the domain size, or roughly the grain size


    In either way of them, a equation with Tc in both right and left hand will be yielded. Then using iterative calculation, the Tc can be determined.

    Parameters
    -------
    da2: D/a2 in 1/a, pear year
    sda2:
    E: diffusion activation energy
    sE:
    pho: covariance of da2 and E
    temp: temperature of the given diffusion coefficient, in K
    A: geometric constant
    R: gas constant
    cooling_rate: in degree/m.y., degrees per million year
    temp_in_celsius: temperature in celsius, True, else in K, False

    Returns
    -------

    """
    R = 8.314 if R is None else R # in J/K mol
    Tc = 600
    sTc = 0
    Tm = 9999999999 if temp is None else temp
    Tm = Tm + 273.15 if temp_in_celsius else Tm
    cooling_rate = cooling_rate / 1000000

    iter_num = 0
    iter_diff = 100
    while abs(iter_diff) > 0.01 and iter_num < 100:
        tau = R * Tc ** 2 / (E * cooling_rate)
        new_Tc = (R / E) * log(A * tau * da2) + 1 / Tm  #
        # new_Tc = (R / E) * log(A * tau * da2)
        d1 = cooling_rate / (A * da2 * Tc ** 2)
        s1 = d1 ** 2 * sda2 ** 2  # da2
        d2 = R / E ** 2 * (log(A * tau * da2) + 1)
        s2 = d2 ** 2 * sE ** 2  # E
        d3 = 2 * R / (E * Tc)
        s3 = d3 ** 2 * sTc ** 2  # Tc
        new_Tc = 1 / new_Tc
        iter_diff = abs(new_Tc - Tc)
        Tc = new_Tc
        sTc = Tc ** 2 * np.sqrt(s1 + s2 + s3 + 2 * d1 * d2 * pho)
        iter_num += 1
        # print(f"Get Tc: {iter_num = }, {Tc = }, {tau = }, {Tm = }")
    return Tc - 273.15 if temp_in_celsius else Tc, sTc


if __name__ == "__main__":
    """
    tests
    """
    # 柴田贤, 钾长石
    print("# 柴田贤, 钾长石")
    print(get_tc(da2=5.6 * 3600 * 24 * 365.24, sda2=0, E=28.8 * 1000, R=1.987, sE=0, cooling_rate=5, A=8.7))
    print(get_tc(da2=5.6 * 3600 * 24 * 365.24, sda2=0, E=28.8 * CAL2JOULE * 1000, sE=0, cooling_rate=5, A=8.7))
    print(get_tc(da2=5.6 * 3600 * 24 * 365.24, sda2=0, E=28.8 * CAL2JOULE * 1000, sE=0, cooling_rate=30, A=8.7))
    # 杨静, 斜钾铁矾 K101-1
    print("# 杨静, 斜钾铁矾 K101-1")
    print(get_tc(da2=10 ** 15.07 * 3600 * 24 * 365.24, sda2=0, E=76.17 * 1000 * CAL2JOULE, sE=0, cooling_rate=5, A=55))  # 这里跟文章中吻合
    print(get_tc(da2=10 ** 15.07 * 3600 * 24 * 365.24, sda2=0, E=76.17 * 1000 * CAL2JOULE, sE=0, cooling_rate=10, A=55))  # 10冷却速率不吻合
    # 杨静, 斜钾铁矾 K101-2
    print("# 杨静, 斜钾铁矾 K101-2")
    print(get_tc(da2=10 ** 12.35 * 3600 * 24 * 365.24, sda2=0, E=66.43 * 1000 * CAL2JOULE, sE=0, cooling_rate=5, A=55))  # 这里跟文章中吻合
    print(get_tc(da2=10 ** 12.35 * 3600 * 24 * 365.24, sda2=0, E=66.43 * 1000 * CAL2JOULE, sE=0, cooling_rate=10, A=55))  # 10冷却速率不吻合
    # 芦武长
    print("# 芦武长")
    print(get_tc(da2=3e-4 * 3600 * 24 * 365.24, sda2=0, E=35 * 1000 * CAL2JOULE, sE=0, cooling_rate=0.1, A=27, temp=600))  # 缺乏样品特征尺寸参数，da2不对，结果不对
    print(get_tc(da2=3e-4 * 3600 * 24 * 365.24, sda2=0, E=35 * 1000 * CAL2JOULE, sE=0, cooling_rate=1, A=27, temp=600))  # 缺乏样品特征尺寸参数，da2不对，结果不对
    # Harrison 1981, 角闪石, A = 55
    print("# Harrison 1981, 角闪石, A = 55")
    print(get_tc(da2=0.024 * SEC2YEAR / 0.0080 ** 2, sda2=0, E=64.1 * 1000 * CAL2JOULE, sE=0, cooling_rate=500, A=55))
    print(get_tc(da2=0.024 * SEC2YEAR / 0.0080 ** 2, sda2=0, E=64.1 * 1000 * CAL2JOULE, sE=0, cooling_rate=5, A=55))
    # Harrison 1985, 黑云母, A = 27
    print("# Harrison 1985, 黑云母, A = 27")
    print(get_tc(da2=0.077 * SEC2YEAR / 0.0150 ** 2, sda2=0, E=47.0 * 1000 * CAL2JOULE, sE=0, cooling_rate=100, A=27))
    print(get_tc(da2=0.077 * SEC2YEAR / 0.0150 ** 2, sda2=0, E=47.0 * 1000 * CAL2JOULE, sE=0, cooling_rate=1, A=27))
    print("# Blereau 2019, 大隅石，Osumilite, A = 8.7")
    print(get_tc(da2=8.34e8 * SEC2YEAR / (0.0175 ** 2), sda2=0, E=461 * 1000, sE=0, cooling_rate=10, A=8.7))
    print(get_tc(da2=6.70e8 * SEC2YEAR / (0.0175 ** 2), sda2=0, E=453 * 1000, sE=0, cooling_rate=10, A=8.7))  # 这两个与原文对不上，且原文描述表格与文字不符
    print(get_tc(da2=4.49e5 * SEC2YEAR / (0.0175 ** 2), sda2=0, E=369 * 1000, sE=0, cooling_rate=10, A=8.7))  # 这两个与原文对不上，且原文描述表格与文字不符
    print("# Thern 2020, 电气石, A = 55")  # 大部分都对或者接近，不完全一致的原因大概是他们采用蒙特卡洛方法计算且正负误差不一样
    print(get_tc(da2=4.86e22 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=678 * 1000, sE=0, cooling_rate=10, A=55))
    print(get_tc(da2=8.99e23 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=756 * 1000, sE=0, cooling_rate=10, A=55))
    print(get_tc(da2=3.95e19 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=604 * 1000, sE=0, cooling_rate=10, A=55))
    print(get_tc(da2=6.95e27 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=815 * 1000, sE=0, cooling_rate=10, A=55))
    print(get_tc(da2=2.70e14 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=519 * 1000, sE=0, cooling_rate=10, A=55))
    print(get_tc(da2=1.65e15 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=505 * 1000, sE=0, cooling_rate=10, A=55))
    print(get_tc(da2=9.09e21 * SEC2YEAR / (0.0100 ** 2), sda2=0, E=656 * 1000, sE=0, cooling_rate=10, A=55))
    print("# Shi 2020, 黑云母, A = 55")  # 大部分都对或者接近，不完全一致的原因大概是他们采用蒙特卡洛方法计算且正负误差不一样
    print(get_tc(da2=10 ** 18 * SEC2YEAR, sda2=0, E=87 * 1000 * CAL2JOULE, sE=0, cooling_rate=10, A=6))  # 文中的参数可能是cooling_rate=100, A=8.7，但是10和6的组合与文中报导的结果一致
    print(get_tc(da2=10 ** 16.6 * SEC2YEAR, sda2=0, E=81.5 * 1000 * CAL2JOULE, sE=0, cooling_rate=10, A=6))
